from setuptools import setup

setup(
  name="epubcfi",
  version="0.0.1",
  author="Tao Zeyu",
  author_email="i@taozeyu.com",
  description="handle Epub CFI",
  packages=["epubcfi"],
  long_description=open("./README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  install_requires=[],
)